#include <strings.h>
#include "watch.h"
#include "m4ri.h"

//good 36586 = 1.338535 billion
//bad  36687 = 1.345936 billion
// 1338 million bits is "safe".

int verbose, silent, dimension, number;
packedmatrix *identitymatrix;

/***************************************************************
 * Inverts a matrix [A] by augmenting it with the identity matrix [E] and
 * performing Gaussian elimination on the resulting matrix [AE].
 *
 * The Gaussian elimintation is timed.
 *
 **/
int experiment(int size) {
  packedmatrix *mymatrix=createPackedMatrix(size, size);
  packedmatrix *combine, *inverse;
  int rank, err=0;

  if (verbose==YES) printf("/***************************************************/\n");
  if (verbose==YES) printf("/***************************************************/\n");
  if ((verbose==NO) && (silent==NO)) printf("Starting to generate matrix.\n");

  fillRandomlyPacked(mymatrix);
  
  if (verbose==YES) printf("The original matrix is:\n");
  if (verbose==YES) lazyPrint(mymatrix);
  if ((verbose==NO) && (silent==NO)) printf("Done generating matrix.\n");

  combine=concatPacked(mymatrix, identitymatrix);
  
  if (verbose==YES) printf("And is augmented to: \n");
  if (verbose==YES) lazyPrint(combine);
  if ((verbose==NO) && (silent==NO)) printf("Starting Gaussian Elmination.\n");

  startWatch();
  rank=gaussianPacked(combine,1);
  stopWatch();

  if (rank != combine->rows) {
    if (err==0) if (silent==NO) printf("No errors.\n");
    if (silent==NO) {
      printf("Darn, singular.\n\n");
      printf("Time was %ld microsec.\n",getWatch());
      if (verbose==YES) printf("And this is how far I got: \n");
      if (verbose==YES) lazyPrint(combine);
    }

    destroyPackedMatrix(combine);
    destroyPackedMatrix(mymatrix);
    return YES;
  }

  printf("Time was %ld microsec.\n",getWatch());
  store(getWatch());

  if (verbose==YES) printf("And this reduces to: \n");
  if (verbose==YES) lazyPrint(combine);
  if ((verbose==NO) && (silent==NO)) printf("Finished with Gaussian Elimination.\n");  

  inverse=copySubMatrixPacked(combine, 0, size, size-1, size*2-1);
  destroyPackedMatrix(combine);
  
  if (verbose==YES) printf("Which stripped is: \n");
  if (verbose==YES) lazyPrint(inverse);

  //destroyPackedMatrix(inverse);
  destroyPackedMatrix(mymatrix);

  if ((silent==NO) && (err==0)) printf("No errors.\n");
  if (silent==NO) printf("Matrix is non-singular.\n\n");

  return NO;
}

/****************************************************************/
void processArgs( int argc, char **argv ) {
  int i;

  verbose=0; silent=0;

  /* printf("There are %d args.\n", argc-1);
  for (i=0; i<argc; i++) {
    printf("Arg %d: %s\n", i, argv[i]);
    }*/

  if (argc<3) die("Usage: %s DIMENSION EXPERIMENTS [-s -v]");

  dimension=atoi(argv[1]);
  number=atoi(argv[2]);

  for (i=3; i<argc; i++) {
    if (strcasecmp(argv[i], "-v")==0) verbose=YES;
    else if (strcasecmp(argv[i], "-s")==0) silent=YES;
    else { printf("Unknown arg: \"%s\"\n", argv[i]); die("Bad Args");}
  }

  printf("Dim=%d, Num=%d, Silent=%d, Verbose=%d.\n", dimension, number, silent, verbose);

  if (dimension<1) die("Bad Dimension.");
  if (number<1) die("Bad Number.");

  if ((verbose==YES) && (silent==YES)) die("Verbose and Silnet are incompatible.\n");
}

/****************************************************************/

int main ( int argc, char **argv) {
  int i;
  //matrix *first;
  int count=0;

  clock();

  processArgs(argc, argv);

#ifndef UNPACKED
  setupPackingMasks();
#endif

  identitymatrix=createPackedMatrix(dimension, dimension);
  makeIdentityPacked(identitymatrix);

  clearLogs();
  seedWatch();
  
  for (i=0; i<number; i++) {
    count+=experiment(dimension);
  }

  printf("Average time: %f\n", getAverage());
  printf("Sigma: %f\n", getSigma());
  printf("Count: %ld\n", getCount());

  printf("\n\n Of %d matrices, %d were singular, or %f%%.\n",
	 number, count, (count*100.0)/(number*1.0));

  destroyPackedMatrix(identitymatrix);

  return 0;
}





