#include <strings.h>
#include "m4ri.h"

//good 36586 = 1.338535 billion
//bad  36687 = 1.345936 billion
// 1338 million bits is "safe".

int verbose, silent, dimension, number;
packedmatrix *identitymatrix;

/****************************************************************/
int experiment(int size) {
  //int err=0;
  packedmatrix *mymatrix;
  
  mymatrix=createPackedMatrix(size,size);
  fillRandomlyPacked(mymatrix);

  printf("Original Matrix:\n");
  lazyPrint(mymatrix);
  simpleFourRussiansPackedFlex(mymatrix,6,YES);
  printf("Final Matrix:\n");
  lazyPrint(mymatrix);

  return NO;
}

/****************************************************************/
void processArgs( int argc, char **argv ) {
  int i;

  verbose=0; silent=0;

  /* printf("There are %d args.\n", argc-1);
  for (i=0; i<argc; i++) {
    printf("Arg %d: %s\n", i, argv[i]);
    }*/

  if (argc<3) die("Usage: %s DIMENSION EXPERIMENTS [-s -v]");

  dimension=atoi(argv[1]);
  number=atoi(argv[2]);

  for (i=3; i<argc; i++) {
    if (strcasecmp(argv[i], "-v")==0) verbose=YES;
    else if (strcasecmp(argv[i], "-s")==0) silent=YES;
    else { printf("Unknown arg: \"%s\"\n", argv[i]); die("Bad Args");}
  }

  printf("Dim=%d, Num=%d, Silent=%d, Verbose=%d.\n", dimension, number, silent, verbose);

  if (dimension<1) die("Bad Dimension.");
  if (number<1) die("Bad Number.");

  if ((verbose==YES) && (silent==YES)) die("Verbose and Silnet are incompatible.\n");
}

/****************************************************************/

int main ( int argc, char **argv) {
  int i;
  //matrix *first;
  int count=0;

  clock();

  processArgs(argc, argv);


  setupPackingMasks(); 
  seedWatch(); 
  buildAllCodes(); 

  identitymatrix=createPackedMatrix(dimension, dimension);
  makeIdentityPacked(identitymatrix);

  clearLogs();

#ifndef UNPACKED
  setupPackingMasks();
#endif

  seedWatch();
  
  for (i=0; i<number; i++) {
    count+=experiment(dimension);
  }

  printf("Average time: %f\n", getAverage());
  printf("Sigma: %f\n", getSigma());
  printf("Count: %ld\n", getCount());

  printf("\n\n Of %d matrices, %d were singular, or %f%%.\n",
	 number, count, (count*100.0)/(number*1.0));

  destroyPackedMatrix(identitymatrix);

  destroyAllCodes();

  return 0;
}





