#include <strings.h>
#include "m4ri.h"

//good 36586 = 1.338535 billion
//bad  36687 = 1.345936 billion
// 1338 million bits is "safe".

int verbose, silent, dimension, number;
matrix *identitymatrix;

double gausstotal, multtotal;

/****************************************************************/
int newexperiment() {
  matrix *mymatrix=createMatrix(dimension, dimension);
  matrix *combine, *inverse, *product;
  int nonsingular;
  clock_t gausstime, multtime;
  int size=dimension;

  printf("/*************************");
  printf("Begining to fill matrix.\n");
  fillRandomMatrix(mymatrix);
  printf("Matrix filled.\n");

  combine=concatMatrix(mymatrix, identitymatrix);

  printf("Starting gaussian elimination.\n");
  startWatch();
  nonsingular=gaussian(combine,1);
  stopWatch();
  printf("Gaussian Elimination complete.\n");
  gausstime=getWatch();

  if (nonsingular==NO) {
    printf("Darn, singular.\n");
    printf("Time was %d ms.\n",(int)getWatch()/1000);
    destroyMatrix(combine); 
    destroyMatrix(mymatrix); 
    return NO;
  }

  inverse=copySubMatrix(combine, 0, size, size-1, size*2-1);

  printf("Starting multiply.\n");
  startWatch();
  product=matrixTimesMatrix(inverse, mymatrix);
  stopWatch();
  printf("Done with multiply.\n");
  multtime=getWatch();

  if (equalMatrix(product, identitymatrix)==NO) {
    printf("ERROR!!!\n");
    destroyMatrix(inverse); destroyMatrix(combine); destroyMatrix(mymatrix); destroyMatrix(product);
    return YES;
  } else {
    printf("Gauss Elim = %d, Multiply = %d.\n", (int)gausstime/1000, (int)multtime/1000);
    gausstotal+=gausstime;
    multtotal+=multtime;
    destroyMatrix(inverse); destroyMatrix(combine); destroyMatrix(mymatrix); destroyMatrix(product);
    return YES;
  }
}
/****************************************************************/
int experiment(int size) {
  matrix *mymatrix=createMatrix(size, size);
  matrix *combine, *inverse = NULL;
  int singular, err=0;

  if (verbose==YES) printf("/***************************************************/\n");
  if (verbose==YES) printf("/***************************************************/\n");
  if ((verbose==NO) && (silent==NO)) printf("Starting to generate matrix.\n");

  fillRandomMatrix(mymatrix);
  
  if (verbose==YES) printf("The original matrix is:\n");
  if (verbose==YES) printMatrix(mymatrix);
  if ((verbose==NO) && (silent==NO)) printf("Done generating matrix.\n");

  combine=concatMatrix(mymatrix, identitymatrix);
  
  if (verbose==YES) printf("And is augmented to: \n");
  if (verbose==YES) printMatrix(combine);
  if ((verbose==NO) && (silent==NO)) printf("Starting Gaussian Elmination.\n");

  startWatch();
  singular=gaussian(combine,1);
  stopWatch();

  if (singular==NO) {
    if (err==0) if (silent==NO) printf("No errors.\n");
    if (silent==NO) {
      printf("Darn, singular.\n\n");
      printf("Time was %ld microsec.\n",getWatch());
      if (verbose==YES) printf("And this is how far I got: \n");
      if (verbose==YES) printMatrix(combine);
    }

    destroyMatrix(combine);
    destroyMatrix(mymatrix);
    return YES;
  }

  printf("Time was %ld microsec.\n",getWatch());
  store(getWatch());

  if (verbose==YES) printf("And this reduces to: \n");
  if (verbose==YES) printMatrix(combine);
  if ((verbose==NO) && (silent==NO)) printf("Finished with Gaussian Elimination.\n");  

  //inverse=copySubMatrix(combine, 0, size, size-1, size*2-1);
  destroyMatrix(combine);
  
  if (verbose==YES) printf("Which stripped is: \n");
  if (verbose==YES) printMatrix(inverse);

  //destroyMatrix(inverse);
  destroyMatrix(mymatrix);

  if ((silent==NO) && (err==0)) printf("No errors.\n");
  if (silent==NO) printf("Matrix is non-singular.\n\n");

  return NO;
}

/****************************************************************/
void processArgs( int argc, char **argv ) {
  int i;

  verbose=0; silent=0;

  /* printf("There are %d args.\n", argc-1);
  for (i=0; i<argc; i++) {
    printf("Arg %d: %s\n", i, argv[i]);
    }*/

  if (argc<3) die("Usage: %s DIMENSION EXPERIMENTS [-s -v]");

  dimension=atoi(argv[1]);
  number=atoi(argv[2]);

  for (i=3; i<argc; i++) {
    if (strcasecmp(argv[i], "-v")==0) verbose=YES;
    else if (strcasecmp(argv[i], "-s")==0) silent=YES;
    else { printf("Unknown arg: \"%s\"\n", argv[i]); die("Bad Args");}
  }

  printf("Dim=%d, Num=%d, Silent=%d, Verbose=%d.\n", dimension, number, silent, verbose);

  if (dimension<1) die("Bad Dimension.");
  if (number<1) die("Bad Number.");

  if ((verbose==YES) && (silent==YES)) die("Verbose and Silnet are incompatible.\n");
}

/****************************************************************/

int main ( int argc, char **argv) {
  int i;
  //matrix *first;
  int count=0;

  clock();

  processArgs(argc, argv);

  setupPackingMasks(); 
  seedWatch(); 
  buildAllCodes(); 

  gausstotal=0;
  multtotal=0;
  identitymatrix=createMatrix(dimension, dimension);
  setIdentityMatrix(identitymatrix);

  clearLogs();
  seedWatch();
  
  for (i=0; i<number; i++) {
    count+=newexperiment();
  }

  printf("Average gaussian time: %f\n", gausstotal/(count*1000.0));
  printf("Average multiply time: %f\n", multtotal/(count*1000.0));
  printf("Count: %ld\n", getCount());

  printf("\n\n Of %d matrices, %d were singular, or %f%%.\n",
	 number, count, (count*100.0)/(number*1.0));

  destroyMatrix(identitymatrix);

  return 0;
}





