#include <strings.h>
#include "m4ri.h"

//good 36586 = 1.338535 billion
//bad  36687 = 1.345936 billion
// 1338 million bits is "safe".

int verbose, silent, dimension, number;
packedmatrix *identitymatrix;

/****************************************************************/
int experiment(int size) {
  int rank, k;
  packedmatrix *mymatrix, *bigmatrix1, *bigmatrix2;
  clock_t  gaussiantime;
 
  printf("/**************************************/\n");
  printf("Starting experiment.\n");

  printf("Allocating.\n");
  mymatrix=createPackedMatrix(size,size);

  printf("Filling.\n");
  fillRandomlyPacked(mymatrix);

  printf("Concating.\n");
  bigmatrix1=concatPacked(mymatrix, identitymatrix);
  bigmatrix2=concatPacked(mymatrix, identitymatrix);

  printf("Executing Gaussian.\n");
  startWatch();
  rank=gaussianPacked(bigmatrix2, YES);
  stopWatch();
  gaussiantime=getWatch();

  printf("The Gaussian alg took %d ms.\n", (int)gaussiantime/1000);

  if (rank != bigmatrix2->rows) {
    printf("Darn, singular.\n");

    destroyPackedMatrix(mymatrix);
    destroyPackedMatrix(bigmatrix1);
    destroyPackedMatrix(bigmatrix2);
    return YES;
  }

  k=2;

  do {
    printf("Starting k=%d.\n", k);
    startWatch();
    simpleFourRussiansPackedFlex(bigmatrix2, k, YES);
    stopWatch();
    printf("That took %d ms.\n", (int)getWatch()/1000);
    k++;
    bigmatrix2=concatPacked(mymatrix, identitymatrix);
  } while (k<11);


  destroyPackedMatrix(mymatrix);
  destroyPackedMatrix(bigmatrix1);
  destroyPackedMatrix(bigmatrix2);
  
  return NO;
}

/****************************************************************/
void processArgs( int argc, char **argv ) {
  int i;

  verbose=0; silent=0;

  /* printf("There are %d args.\n", argc-1);
  for (i=0; i<argc; i++) {
    printf("Arg %d: %s\n", i, argv[i]);
    }*/

  if (argc<3) die("Usage: %s DIMENSION EXPERIMENTS [-s -v]");

  dimension=atoi(argv[1]);
  number=atoi(argv[2]);

  for (i=3; i<argc; i++) {
    if (strcasecmp(argv[i], "-v")==0) verbose=YES;
    else if (strcasecmp(argv[i], "-s")==0) silent=YES;
    else { printf("Unknown arg: \"%s\"\n", argv[i]); die("Bad Args");}
  }

  printf("Dim=%d, Num=%d, Silent=%d, Verbose=%d.\n", dimension, number, silent, verbose);

  if (dimension<1) die("Bad Dimension.");
  if (number<1) die("Bad Number.");

  if ((verbose==YES) && (silent==YES)) die("Verbose and Silnet are incompatible.\n");
}

/****************************************************************/

int main ( int argc, char **argv) {
  int i;
  //matrix *first;
  int count=0;

  clock();

  processArgs(argc, argv);

  setupPackingMasks(); 
  seedWatch(); 
  buildAllCodes(); 

  identitymatrix=createPackedMatrix(dimension, dimension);
  makeIdentityPacked(identitymatrix);

  clearLogs();


  seedWatch();
  
  for (i=0; i<number; i++) {
    count+=experiment(dimension);
  }

  destroyPackedMatrix(identitymatrix);

  return 0;
}





