#include <strings.h>
#include "m4ri.h"

//good 36586 = 1.338535 billion
//bad  36687 = 1.345936 billion
// 1338 million bits is "safe".

int verbose, silent, dimension, number;

/***************************************************************
 * Inverts a matrix [A] by augmenting it with the identity matrix [E] and
 * performing Gaussian elimination on the resulting matrix [AE].
 *
 * The result is checked by left and right multiplying the inverted
 * matrix B = A^-1 with A. 
 *
 **/


int experiment(int size) {
  packedmatrix *left=createPackedMatrix(size, size);
  packedmatrix *right=createPackedMatrix(size, size);
  packedmatrix *res1, *res2;
  clock_t naivtime, m4rmtime;

  if (verbose==YES) printf("/***************************************************/\n");
  if (verbose==YES) printf("/***************************************************/\n");
  if ((verbose==NO) && (silent==NO)) printf("Starting to generate matrix.\n");

  fillRandomlyPacked(left);
  fillRandomlyPacked(right);

  /*  makeIdentityPacked(left); */
  /*  makeIdentityPacked(right); */

  
  if (verbose==YES) printf("The left matrix is:\n");
  if (verbose==YES) lazyPrint(left);
  if (verbose==YES) printf("The right matrix is:\n");
  if (verbose==YES) lazyPrint(right);
  if ((verbose==NO) && (silent==NO)) printf("Done generating matrix.\n");


  printf("NAIV\n");
  startWatch();
  //res1 = matrixTimesMatrixPacked(left,right);
  res1 = m4rmTransposePacked(left,right, 11);
  stopWatch();
  naivtime=getWatch();

  printf("M4RM\n");
  startWatch();
  res2 = m4rmPacked(left,right, 11);
  stopWatch();
  m4rmtime=getWatch();

  if (verbose==YES) printf("The naiv result is:\n");
  if (verbose==YES) lazyPrint(res1);


  if (verbose==YES) printf("The M4RM result is:\n");
  if (verbose==YES) lazyPrint(res2);

  if(equalPackedMatrix(res1,res2)) {
    printf("matrices match\n");
  } else {
    die("matrices don't match");
  }

  printf("n^3 Time = %d, M4RM Time = %d.\n", (int)naivtime/1000, (int)m4rmtime/1000);

  destroyPackedMatrix(left);
  destroyPackedMatrix(right);
  destroyPackedMatrix(res1);
  destroyPackedMatrix(res2);

  return NO;
}

/****************************************************************/
void processArgs( int argc, char **argv ) {
  int i;

  verbose=0; silent=0;

  /* printf("There are %d args.\n", argc-1);
  for (i=0; i<argc; i++) {
    printf("Arg %d: %s\n", i, argv[i]);
    }*/

  if (argc<3) die("Usage: %s DIMENSION EXPERIMENTS [-s -v]");

  dimension=atoi(argv[1]);
  number=atoi(argv[2]);

  for (i=3; i<argc; i++) {
    if (strcasecmp(argv[i], "-v")==0) verbose=YES;
    else if (strcasecmp(argv[i], "-s")==0) silent=YES;
    else { printf("Unknown arg: \"%s\"\n", argv[i]); die("Bad Args");}
  }

  printf("Dim=%d, Num=%d, Silent=%d, Verbose=%d.\n", dimension, number, silent, verbose);

  if (dimension<1) die("Bad Dimension.");
  if (number<1) die("Bad Number.");

  if ((verbose==YES) && (silent==YES)) die("Verbose and Silnet are incompatible.\n");
}

/****************************************************************/

int main ( int argc, char **argv) {
  int i;

  processArgs(argc, argv);

  setupPackingMasks();
  seedWatch(); 
  buildAllCodes(); 

  srandom(5);

  for (i=0; i<number; i++) {
    experiment(dimension);
  }


  return 0;
}


